﻿/*
    Copyright 2009 Vassilis Varveropoulos (vassilis@vrhome.net)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace IMUDisplay
{
    class IMUData
    {
        public const int MaxNumBytes = 500;

        public enum FrameTypeEnum {RT = 0, Pitch, Roll, Angle, IMU, None};
        
        public FrameTypeEnum frameType;     
        public List<IMUDataEntry> entries;
        public int numEntries;
        private object locker = new object();
        private List<IMUDataEntry> workingEntries;
        private float time = -1.0f;

        public IMUData()
        {
            workingEntries = new List<IMUDataEntry>();
            frameType = FrameTypeEnum.None;
            SetFrameType(FrameTypeEnum.RT);
            entries = new List<IMUDataEntry>(workingEntries);
        }

        public int GetNumBytes()
        {
            lock (locker)
            {
                return numEntries * 4;
            }
        }

        public void Update(byte[] data)
        {
            lock (locker)
            {
                // Read each data channels
                int offset = 0;
                for (int n = 0; n < numEntries; ++n)
                {
                    workingEntries[n].Value = BitConverter.ToSingle(data, offset);
                    offset += 4;
                }

                // Update calculated fields
                //Set("AccelPitchAngle", (float)(Math.Atan2(Get("xAxis"), Math.Sqrt(Get("yAxis") * Get("yAxis") + Get("zAxis") * Get("zAxis"))) * 180.0 / Math.PI));
                //Set("AccelRollAngle", (float)(Math.Atan2(Get("yAxis"), Math.Sqrt(Get("xAxis") * Get("xAxis") + Get("zAxis") * Get("zAxis"))) * 180.0 / Math.PI));
                
                // If clock is not updated then update manually 
                if (Get("Clock") == 0 || time > 0)
                {
                    if (time < 0)
                        time = 0.0f;
                    time = time + 0.010f;
                    Set("Clock", (float)time);
                }

                // Update master entry list
                syncEntries();
            }
        }

        public void Update(string line)
        {
            string[] strData = line.Split(' ');
            lock (locker)
            {
                // Check that the data buffer is of the correct size
                if (strData.Length < numEntries)
                    throw new ArgumentException("Invalid line size");

                // Read each data channels
                for (int n = 0; n < numEntries; ++n)
                {
                    workingEntries[n].Value = (float)double.Parse(strData[n]);
                }

                // Update master entry list
                syncEntries();
            }
        }

        public FrameTypeEnum GetFrameType()
        {
            lock (locker)
            {
                return frameType;
            }
        }

        public void SetFrameType(FrameTypeEnum _frameType)
        {
            lock (locker)
            {
                // Update only if type has changed
                if (frameType != _frameType)
                {
                    // Create data entries and add them to a list
                    workingEntries.Clear();
                    frameType = _frameType;
                    if (frameType == FrameTypeEnum.RT)
                    {
                        workingEntries.Add(new IMUDataEntry("xAxis", ""));
                        workingEntries.Add(new IMUDataEntry("yAxis", ""));
                        workingEntries.Add(new IMUDataEntry("zAxis", ""));
                        workingEntries.Add(new IMUDataEntry("xRate", ""));
                        workingEntries.Add(new IMUDataEntry("yRate", ""));
                        workingEntries.Add(new IMUDataEntry("PitchAngle", ""));
                        workingEntries.Add(new IMUDataEntry("RollAngle", ""));
                        workingEntries.Add(new IMUDataEntry("PitchRate", ""));
                        workingEntries.Add(new IMUDataEntry("RollRate", ""));
                        workingEntries.Add(new IMUDataEntry("GainIn", ""));
                        workingEntries.Add(new IMUDataEntry("ElevatorIn", ""));
                        workingEntries.Add(new IMUDataEntry("AileronsIn", ""));
                        workingEntries.Add(new IMUDataEntry("RudderIn", ""));
                        workingEntries.Add(new IMUDataEntry("ThrottleIn", ""));
                        workingEntries.Add(new IMUDataEntry("PitchCommand", ""));
                        workingEntries.Add(new IMUDataEntry("RollCommand", ""));
                        workingEntries.Add(new IMUDataEntry("YawCommand", ""));
                        workingEntries.Add(new IMUDataEntry("ThrottleCommand", ""));
                        workingEntries.Add(new IMUDataEntry("StabEnabled", ""));
                        workingEntries.Add(new IMUDataEntry("LoopRunTime", ""));
                        workingEntries.Add(new IMUDataEntry("IMURunTime", ""));
                        workingEntries.Add(new IMUDataEntry("GainSetpoint", ""));
                        workingEntries.Add(new IMUDataEntry("PitchAngleSetpoint", ""));
                        workingEntries.Add(new IMUDataEntry("RollAngleSetpoint", ""));
                        workingEntries.Add(new IMUDataEntry("Debug", ""));
                        workingEntries.Add(new IMUDataEntry("ErrorCode", ""));
                        workingEntries.Add(new IMUDataEntry("Clock", ""));
                        workingEntries.Add(new IMUDataEntry("State", ""));
                        workingEntries.Add(new IMUDataEntry("MotorFrontCommand", ""));
                        workingEntries.Add(new IMUDataEntry("MotorRearCommand", ""));
                        workingEntries.Add(new IMUDataEntry("MotorRightCommand", ""));
                        workingEntries.Add(new IMUDataEntry("MotorLeftCommand", ""));
                        numEntries = workingEntries.Count;

                    }
                    else if (frameType == FrameTypeEnum.Angle)
                    {
                        workingEntries.Add(new IMUDataEntry("PitchAngle", ""));
                        workingEntries.Add(new IMUDataEntry("RollAngle", ""));
                        numEntries = workingEntries.Count;
                    }
                    else
                    {
                        throw new ArgumentException("Invalid frame type");
                    }
                    // Add common entries
                    workingEntries.Add(new IMUDataEntry("Connected", ""));               
                }
            }
        }


        public float Get(string name)
        {
            lock (locker)
            {
                // Find requested name
                for (int n = 0; n < entries.Count; ++n)
                {
                    if (entries[n].Name.Equals(name))
                        return entries[n].Value;
                }
                // If this point is reached name was not found
                throw new ArgumentException();
            }
        }

        public void Set(string name, float val)
        {
            lock (locker)
            {
                // Find requested name (main list)
                for (int n = 0; n < entries.Count; ++n)
                {
                    if (entries[n].Name.Equals(name))
                    {
                        entries[n].Value = val;
                    }
                }
                // Find requested name (working list)
                for (int n = 0; n < workingEntries.Count; ++n)
                {
                    if (workingEntries[n].Name.Equals(name))
                    {
                        workingEntries[n].Value = val;
                    }
                }
            }
        }

        private void syncEntries()
        {
            for (int n = 0; n < entries.Count; ++n)
            {
                for (int m = 0; m < workingEntries.Count; ++m)
                {
                    if (entries[n].Name.Equals(workingEntries[m].Name))
                    {
                        entries[n].Value = workingEntries[m].Value;
                    }
                }
            }
        }
    }

    public class IMUDataEntry
    {
        private string name;
        private float val;
        private string unit;

        public IMUDataEntry(string _name, string _unit)
        {
            name = _name;
            val = 0.0f;
            unit = _unit;
        }

        public string Name
        {
            get { return name; }
            set { name = value; }
        }

        public float Value
        {
            get { return val; }
            set { val = value; }
        }

        public string Unit
        {
            get { return unit; }
            set { unit = value; }
        }
    }
}
