﻿/*
    Copyright 2009 Vassilis Varveropoulos (vassilis@vrhome.net)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.IO;
using NPlot;

namespace IMUDisplay
{
    public partial class MainForm : Form
    {
        private const long MinUpdatePeriod = 50; // ms

        private enum FrameTypeEnum { RT = 0, Angle = 1 };

        private IMUData imuData = new IMUData();
        private string recFilename = null;
        private bool recEnabled = false;
        private bool playbackEnabled = false;
        private string cmdStr = "a";
        private DateTime lastUpdateTime;
        private FrameTypeEnum frameType = FrameTypeEnum.RT;

        public MainForm()
        {
            InitializeComponent();

            dataGrid.DataSource = imuData.entries;
            imuData.Set("Connected", 0);
            dataGrid.Refresh();
            dataGrid.AutoResizeColumns();
            frameTypeCb.SelectedIndex = (int)(frameType);
            commsWorker.RunWorkerAsync();
        }

        private void worker_DoWork(object sender, DoWorkEventArgs e)
        {
            byte[] data = new byte[IMUData.MaxNumBytes];
            TextWriter recFile = null;
            TextReader playbackFile = null;
            float lastTime = -1.0f;

            // Initialize serial port 
            serialPort.BaudRate = 19200;
            serialPort.PortName = "COM5";
            serialPort.ReadTimeout = 500;
            serialPort.Open();

            // Read loop
            while (!commsWorker.CancellationPending)
            {
                if (!playbackEnabled)
                {
                    try
                    {
                        // Close playback file if not already done
                        if (playbackFile != null)
                        {
                            playbackFile.Close();
                            playbackFile = null;
                            lastTime = -1.0f;
                        }
                        // Wait for sync 
                        string sync = serialPort.ReadTo("imud");
                        // Read and set frame type
                        int frameType = (int)serialPort.ReadByte();
                        imuData.SetFrameType((IMUData.FrameTypeEnum)(frameType));
                        // Read data packet
                        for (int n = 0; n < imuData.GetNumBytes(); ++n)
                            data[n] = (byte)serialPort.ReadByte();
                        imuData.Set("Connected", 1);
                        // Update data
                        imuData.Update(data);
                        // Send update
                        commsWorker.ReportProgress(0, imuData);
                        // Write to file if enabled
                        if (recEnabled)
                        {
                            // Open file if not done already
                            if (recFile == null)
                            {
                                recFile = new StreamWriter(recFilename, false);
                            }
                            // Write line
                            for (int n = 0; n < imuData.entries.Count; ++n)
                            {
                                recFile.Write(imuData.entries[n].Value);
                                recFile.Write(" ");
                            }
                            recFile.WriteLine();
                        }
                        else
                        {
                            // Close file if not done already
                            if (recFile != null)
                            {
                                recFile.Close();
                                recFile = null;
                            }
                        }
                    }
                    catch (TimeoutException ex)
                    {
                        imuData.Set("Connected", 0);
                        serialPort.Write(cmdStr);
                        commsWorker.ReportProgress(0, imuData);
                    }
                }
                else
                {
                    // Close recording file if it is open
                    if (recFile != null)
                    {
                        recFile.Close();
                        recFile = null;
                    }
                    // Open playback file if not already open
                    if (playbackFile == null)
                    {
                        playbackFile = new StreamReader(recFilename);
                        imuData.SetFrameType(IMUData.FrameTypeEnum.RT);
                    }
                    // Read next line from the file 
                    string line = playbackFile.ReadLine();
                    if (line != null)
                    {
                        // Update data
                        imuData.Update(line);
                        // Wait until the time in the update is reached
                        if (lastTime > 0.0f)
                        {
                            float delayTime = imuData.Get("Clock") - lastTime;
                            Thread.Sleep((int)Math.Round(delayTime * 1000.0));
                        }
                        lastTime = imuData.Get("Clock");
                        // Send update
                        commsWorker.ReportProgress(0, imuData);
                    }
                    else
                    {
                        playbackEnabled = false;
                    }
                }
            }

            // Close port
            //serialPort.Close();
        }

        private void worker_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            IMUData imu = (e.UserState as IMUData);
            // Limit update rate of the UI
            DateTime timeNow = DateTime.Now;
            TimeSpan delaySinceLastUpdate = timeNow - lastUpdateTime;
            if (delaySinceLastUpdate.TotalMilliseconds < MinUpdatePeriod)
                return;
            else
                lastUpdateTime = timeNow;
            // Update horizon
            imuHorizon.pitch_angle = imu.Get("PitchAngle");
            imuHorizon.roll_angle = imu.Get("RollAngle");
            // Update data table
            //dataGrid.Refresh();
            // Update bars
            motorFrontBar.Value = (int)imu.Get("MotorFrontCommand");
            motorRearBar.Value = (int)imu.Get("MotorRearCommand");
            motorRightBar.Value = (int)imu.Get("MotorRightCommand");
            motorLeftBar.Value = (int)imu.Get("MotorLeftCommand");
            gainSpBar.Value = (int)imu.Get("GainIn");
            yawCmdHBar.Value = (int)imu.Get("YawCommand") - 500;
            yawCmdLBar.Value = 500 - (int)imu.Get("YawCommand");
            pitchCmdLBar.Value = (int)imu.Get("PitchCommand") - 500;
            pitchCmdHBar.Value = 500 - (int)imu.Get("PitchCommand");
            pitchSpLBar.Value = (int)imu.Get("PitchAngleSetpoint");
            pitchSpHBar.Value = -(int)imu.Get("PitchAngleSetpoint");
            rollCmdLBar.Value = (int)imu.Get("RollCommand") - 500;
            rollCmdHBar.Value = 500 - (int)imu.Get("RollCommand");
            rollSpHBar.Value = (int)imu.Get("RollAngleSetpoint");
            rollSpLBar.Value = -(int)imu.Get("RollAngleSetpoint");
            // Update status bar
            if (imu.Get("Connected") == 1.0)
            {
                connectedStatusLb.Text = "Connected";
                connectedStatusLb.BackColor = Color.Green;
            }
            else
            {
                connectedStatusLb.Text = "Not Connected";
                connectedStatusLb.BackColor = Color.Red;
            }
            // Update time label
            imuTimeLb.Text = String.Format("{0} {1:#.0} {2}", "IMU Time:", imu.Get("Clock"), "sec");
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            commsWorker.CancelAsync();
        }

        private void timer_Tick(object sender, EventArgs e)
        {
            motorFrontBar.Refresh();
            motorRearBar.Refresh();
            motorRightBar.Refresh();
            motorLeftBar.Refresh();
            gainSpBar.Refresh();
            yawCmdHBar.Refresh();
            yawCmdLBar.Refresh();
            pitchCmdLBar.Refresh();
            pitchCmdHBar.Refresh();
            pitchSpLBar.Refresh();
            pitchSpHBar.Refresh();
            rollCmdLBar.Refresh();
            rollCmdHBar.Refresh();
            rollSpLBar.Refresh();
            rollSpHBar.Refresh();
            dataGrid.Refresh();
        }

        private void startRecordingToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // Prompt user for file name and enable recording
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.FileName = recFilename;
            dlg.AddExtension= true;
            dlg.CheckFileExists = false;
            dlg.DefaultExt = "txt";
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                recFilename = dlg.FileName;
                recEnabled = true;
                playbackEnabled = false;
                recStatusLb.Text = "Recording";
                recStatusLb.BackColor = Color.Red;
            }
        }

        private void stopRecordingToolStripMenuItem_Click(object sender, EventArgs e)
        {
            recEnabled = false;
            recStatusLb.Text = "Not Recording";
            recStatusLb.BackColor = Color.Green;
        }

        private void startPlaybackToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // Prompt user for file name and enable playback
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.FileName = recFilename;
            dlg.AddExtension= true;
            dlg.CheckFileExists = false;
            dlg.DefaultExt = "txt";
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                recFilename = dlg.FileName;
                recEnabled = false;
                playbackEnabled = true;
                recStatusLb.Text = "Playback";
                recStatusLb.BackColor = Color.Orange;
            }
        }

        private void stopPlaybackToolStripMenuItem_Click(object sender, EventArgs e)
        {
            playbackEnabled = false;
            recStatusLb.Text = "Not Recording";
            recStatusLb.BackColor = Color.Green;        
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void frameTypeCb_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (frameTypeCb.SelectedIndex == 0)
                cmdStr = "r";
            else if (frameTypeCb.SelectedIndex == 1)
                cmdStr = "a";
        }

    }

}
